; multimode programmable four-wheel-steering controller
; controls two servos from one steering input plus one modeswitch channel
;
; copyright (c) 2008 Alexander Zangerl <az@snafu.priv.at>
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation (version 2 of the License).
	
; $Id: fourws.asm,v 1.11 2010/04/29 13:31:31 az Exp $

; to build for a 12f635: no extra symbols required
; for a 12f683:	 define P12F683
IFDEF	P12F683
processor p12f683
#include "p12f683.inc"
; some renamed register fields
; 635 vs. 683
NOT_RAPU	equ	NOT_GPPU
WPUDA	equ	WPU
RAIE	equ	GPIE
RAIF	equ	GPIF

ELSE
processor p12f635
#include "p12f635.inc"
ENDIF

	RADIX hex
	ERRORLEVEL -302		; ignore the bank warnings, i know what i'm doing :-)

; default speed 4mhz, no clock out, no bod, no copyprotection, no watchdog, no wake-up reset
IFDEF P12F683
; 683 has no wakeup reset
	__CONFIG _INTRC_OSC_NOCLKOUT & _WDT_OFF & _PWRTE_ON & _MCLRE_OFF & _CP_OFF & _CPD_OFF & _BOD_OFF & _IESO_OFF & _FCMEN_OFF
ELSE	
	__CONFIG _INTRC_OSC_NOCLKOUT & _WDT_OFF & _PWRTE_ON & _MCLRE_OFF & _CP_OFF & _CPD_OFF & _BOD_OFF & _IESO_OFF & _FCMEN_OFF & _WUREN_OFF
ENDIF
	
; defaults: mode and min and max acceptable pulse lengths
MAXPULSE	equ	.2200	; officially 2120us
MINPULSE	equ	.840	; officially 920us
CTRPULSE	equ	.1520	; us, official futaba value, jr say 1500
DEADBAND	equ	.160	; us
PROGMINP	equ	.1800	; us. steer and mode must be above that to enter prog mode

MAXDELTA	equ	5	; max # modecheck cycles (of 50ms) between transitions for double transition mode

; i/o pins
STEER	equ	3		; gp3 steering pulses in (in-only anyway)
MSWITCH	equ	4		; gp4 modeswitch pulses in	
FRONT	equ	2		; front steering out
REAR	equ	1		; rear
BEEP	equ	5		; buzzer
LSWITCH equ	0		; fws/rear lockout jumper

; config word bits
REVBIT	equ	0		; 1 reverse rear channel, 0 f+r get same signal
CYCBIT	equ	1		; 1 cyclic mode, 0 direct mode	
DM0BIT	equ	2		; direct mode assignments
DM1BIT	equ	3
RSCBIT	equ	4		; 1 rws is included in cycle
CRCBIT	equ	5		; 1 crab is included in cycle
TRNBIT	equ	6		; 1 cyclic on quick double transition, 0 on hi-lo

; mstate bits
VALBIT	equ	0		; signal within ok range
HIBIT	equ	1		; signal above center
DBBIT	equ	2		; signal within deadband
	

IFDEF	P12F683
VARSTART equ 0x20
ELSE
VARSTART equ 0x40
ENDIF
	cblock VARSTART
	scratch			; tempvar for delay routines

	foo			; tempvars
	bar
	baz
	
	ifoo			; inthandler tempvars
	ibar
	lateneutral
	
	opmode			; 1 4ws, 2 fws, 3 rws, 4 crab
	config			; configuration options, bit-encoded
	newconf	
	lastpins		; to detect pin_changes_
	
	mswitch			; modeswitch, pulse length in multiples of 16us. set by inthandler, read by decodestate
	mstate			; binary state of modeswitch
	prevmstate		; last binary state of modeswitch
	transcount		; signal counter for delta between transitions

	
	neutlo			; steering pulse length in us for neutral
	neuthi
	steerlo			; current steering pulse length in us
	steerhi
	endc

        cblock  0x70            ; bank-shared vars
        w_save                  ; interrupt context
        fsr_save
        status_save
	endc

; i dislike PAGESEL
BANK0	macro
	bcf	STATUS,RP0
	endm
	
BANK1	macro
	bsf	STATUS,RP0
	endm

	org 0			; power-on or reset
	goto main
	org 0x4			; interrupt vector
	movwf   w_save          ; save context
	swapf   STATUS,W
	BANK0
        movwf   status_save
        movfw   FSR
        movwf   fsr_save

	; can safely expect to run in bank0
        call    inthandler

        movfw   fsr_save        
        movwf   FSR
        swapf   status_save,W
        movwf   STATUS
        swapf   w_save,F
        swapf   w_save,W
        retfie

; put the tables in low mem to avoid dealing with pclath and pcl arithmetic
	
; DM1BIT,DM0BIT affect direct modes, selected by hi-neutral-lo on mswitch:
; 00	4ws fws rws
; 01	4ws rws fws
; 10	4ws crab fws
; 11	4ws fws crab
; uses nibbles (hi when dm0 is set), arranged hi-n-lo
; two sets, second used when dm1 is set.
d2mtable:
	addwf	PCL,F
	; dm1 clear
	dt	0x11		; hi, always 4ws
	dt	0x32		; n
	dt	0x23		; lo
	; dm1 set
	dt	0x11		; hi, always 4ws
	dt	0x24		; n
	dt	0x42		; lo
	
; what channels get the direct signal?
; get the unchanged signal in which modes
; returns 0 if no direct outputs at all
directmodes:
	movfw	opmode
	btfsc	config,REVBIT	; reverse rear?
	addlw	4
	addwf	PCL,F
	; the non-reversed settings
	dt	0		; non-mode 0
	dt	(1<<FRONT|1<<REAR) ; mode 1 4ws
	dt	1<<FRONT	; mode 2 2ws-front
	dt	1<<REAR		; mode 3 2ws-rear
	dt	1<<FRONT	; mode 4 crab
	; the reversed settings, same order
	dt	1<<FRONT
	dt	1<<FRONT
	dt	0
	dt	(1<<FRONT|1<<REAR)

; which channels get an inverted signal
; generated on falling edge of orig steering signal
; returns 0 if no inverted signal required
revmodes:
	movfw	opmode
	btfsc	config,REVBIT
	addlw	4
	addwf	PCL,F
	; non-reversed
	dt	0		; non-mode 0
	dt	0		; mode 1 4ws
	dt	0		; mode 2 2ws-front
	dt	0		; mode 3 2ws-rear
	dt	1<<REAR		; mode 4 crab
	; reversed
	dt	1<<REAR		; mode 1 4ws
	dt	0		; mode 2 2ws-front
	dt	1<<REAR		; mode 3 2ws-rear
	dt	0		; mode 4 crab

; which channels get a neutral signal
; generated after falling edge of orig signal and after inverted signal (if required)
; note:	 no difference between reversed and normal mode here.
; returns 0 if no neutral signal required
neutmodes:
	movfw	opmode
	addwf	PCL,F
	dt	0		; non-mode 0
	dt	0		; mode 1 4ws
	dt	1<<REAR		; mode 2 2ws-front
	dt	1<<FRONT	; mode 3 2ws-rear
	dt	0		; mode 4 crab
	

; let's set the environment up properly
main:	movlw	7
	movwf	CMCON0		; no comparator

	BANK1
	movlw	~(1<<FRONT|1<<REAR|1<<BEEP)
	movwf	TRISIO		; those bits output, rest inputs
	movlw	(1<<STEER|1<<MSWITCH)
	movwf	IOCA		; on the r/c inputs we want to get an int-on-change

	; want prescaler for tmr0
	; PS0..2 defaults to 111, want 011 for 1:16 prescale tmr0
	; tmr0 is not yet on, and we don't do overflow ints for it
	; also clear rapu bit for pull-up on lockup pin
	movlw	~(1<<NOT_RAPU|1<<PSA|1<<PS2)
	andwf	OPTION_REG,F

	; note:	on 683 wpuda known as wpu
	clrf	WPUDA		; no pullups except...
	bsf	WPUDA,LSWITCH	; pull up on mode locking pin

IFDEF	P12F683
	clrf	ANSEL		; 683: no analog inputs, please!
ELSE	
	bsf	WDA,LSWITCH	; 635: pull up, not down.
ENDIF
	BANK0

	bsf	INTCON,RAIE	; enable the on-change int 
	bsf	INTCON,PEIE	; and peripheral ints (for tmr1) on but not global yet

	; setup output
	clrf	GPIO
	movfw	GPIO		; read to trigger int-on-change logic

	clrf	neutlo		; a clean slate please.
	clrf	neuthi	
	clrf	steerlo		
	clrf	steerhi
	clrf	mswitch
	clrf	prevmstate
	clrf	transcount

	BANK1			; read config from eeprom
	clrf	EEADR		; configword is at offset 0
	bsf	EECON1,RD
	movfw	EEDATA
	BANK0
	movwf	config

	movlw	1		; normal: start in 4ws mode until calibrated
	btfsc	config,REVBIT	; in reversed-rear mode: all modes but 4 need to reverse and/or
				; neutral, which we don't have yet!
	movlw	4		; hence mode 4
	movwf	opmode		

	call	blip		; hello!
	bsf	INTCON,GIE	; ints on	
	
	; wait a bit for the first few pulses	
moreplease:	
	call	delay100ms

	; check for steering and mode both high, if so enter progmode
	call	decodestate
	movwf	mstate
	
	btfss	mstate,VALBIT	; dud signal? wait a little longer
	goto	moreplease
	btfss	mstate,HIBIT	; skip if mode is high
	goto	docal

	; now check steering, high or not?
	movlw	HIGH PROGMINP
	subwf	steerhi,W	; steer-min should be positive
	skpnz			; hi steer-min = 0? must check lo
	goto	sminlo
	skpc			; carry set = still positive
	goto	docal		; below min
	goto	doprogram	; high is already different, low irrelevant
sminlo:	movlw	LOW PROGMINP
	subwf	steerlo,W	; steer-min should be positive
	skpc			; carry set = still positive
	goto	docal		; below min

doprogram:
	; switch the servos completely off
	clrf	opmode
	
	movlw	1
	movwf	foo		; question counter
	clrf	newconf

	; signal program mode:	8 blips with 256ms pause, then wait 1s and get going.
	movlw	8
	movwf	bar
	call	blip
	clrw
	call	delayXms
	decfsz	bar,F
	goto	$-4
	call	wait1s		; uses bar!

nextquestion:	
	; n long beeps, n is question number
	movfw	foo
	movwf	bar
morebleep:	
	call	bleep
	clrw
	call	delayXms
	decfsz	bar,F
	goto	morebleep

	; extract current setting for this question
	bcf	STATUS,C
	rrf	config,F	; lowest bit now in C
	rrf	newconf,F	; and now in highest
	call	delay100ms	; wait a little until we accept requests
	
	; wait for mode hi->lo transition
wait:	call	decodestate
	movwf	mstate
	; a dud signal for either this or the last just makes us try again
	btfsc	mstate,VALBIT	
	btfss	prevmstate,VALBIT
	goto	morewait
	; so, do we have previous=hi and now=low?
	btfss	mstate,HIBIT	; skip if mstate is high
	btfss	prevmstate,HIBIT; skip if mstate was high before
	goto	morewait
	; wohee, an answer.
	goto	donewait
morewait:
	movfw	mstate		; move mstate to previous
	movwf	prevmstate
	movlw	.20
	call	delayXms
	goto	wait		; and try again
	
donewait:
	movfw	mstate		; move mstate to previous
	movwf	prevmstate
	; capture steering now, low means OFF, high means ON
	; slightly rough, as we miss changes if steering is overwritten right now
	movlw	HIGH CTRPULSE
	subwf	steerhi,W	; steer-neut should be positive for on
	skpnz			; hi steer-neut = 0? must check lo
	goto	checklo
	skpc			; carry set = still positive
	goto	wantzero	; below neut
	goto	wantone		; hi already above neutral, lo irrelevant
checklo:
	movlw	LOW CTRPULSE
	subwf	steerlo,W	; steer-neut should be positive for on
	skpc			; carry set = still positive
	goto	wantzero	; below neut
wantone:
	bsf	newconf,7
	movlw	1
	goto	next
wantzero:
	bcf	newconf,7
	clrw

	; sound new setting, sits in w
next:	call	soundstate	; uses bar, nukes w
	call	wait1s		; and wait one second until you continue
	; next question
	incf	foo,F
	movlw	8		; 7 questions are to be asked
	subwf	foo,W
	skpz
	goto	nextquestion

	; 7 questions: top-most 7 bits of newconf hold that
	; one lower bit is zero and needs to go
	bcf	STATUS,C
	rrf	newconf,F	

	; write the new config word to eeprom
	movfw	newconf
	BANK1
	clrf	EEADR		; statusword at offset 0
	movwf	EEDATA
	bsf	EECON1,WREN
	BCF	INTCON,GIE	; fixed sequence
	movlw	0x55
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1,WR	; end sequence
	BANK0
	
	; sound done pattern: blip every 250ms, forever.
done:	call	blip
	clrw
	call	delayXms
	goto	done

; uses bar, nukes w
wait1s:	movlw	4
	movwf	bar
	clrw
	call	delayXms
	decfsz	bar,F
	goto	$-3
	return

	; calibrate neutral in 8 runs with 250ms delay each
docal:	movlw	8
	movwf	foo

morecal:	movlw	.250
	call	delayXms
	; 16b-add steering to neutral, works fine as max range is 2120
	; and we have 65k
	movfw	steerlo
	addwf	neutlo,F
	movfw	steerhi
	skpnc	
	incf	steerhi,W	; skipped if no carry
	addwf	neuthi,F	; no carry in high byte
	decfsz	foo,F
	goto	morecal

	; now divide neutral by 8 (=3 rightshifts)
	movlw	3
	movwf	foo
dodiv:	bcf	STATUS,C
	rrf	neuthi,F	
	rrf	neutlo,F	; carry from hi moves down into lo
	decfsz	foo,F
	goto dodiv


	btfsc  	T1CON,TMR1ON	; wait until tmr1 is off
	goto	$-1
	; now check the calibration: too high or too low?
	movlw	HIGH MAXPULSE
	subwf	neuthi,W	; neut-max should be negative
	skpnz			; hi neut-max = 0? must check low byte
	goto	cmaxlo
	skpnc			; carry set = still positive
	goto	fault
	goto	cmin		; high is already different, low irrelevant
cmaxlo:	movlw	LOW MAXPULSE
	subwf	neutlo,W	; neut-max should be negative
	skpnc			; carry set = still positive
	goto	fault
cmin:	movlw	HIGH MINPULSE
	subwf	neuthi,W	; neut-min should be positive
	skpnz			; hi neut-min = 0? must check lo
	goto	cminlo
	skpc			; carry set = still positive
	goto	fault
	goto	calok		; high is already different, low irrelevant
cminlo:	movlw	LOW MINPULSE
	subwf	neutlo,W	; neut-min should be positive
	skpc			; carry set = still positive
	goto	fault
	
; one long beep, then show detected neutral for ~2s (100x~20ms)
calok:	call	bleep
	
	movlw	.100		; show for 100x20ms = ~2s
	movwf	foo
	bcf	INTCON,RAIE	; disable pinchange int for the showing part
	clrf	opmode		; no normal outputs!
	
doshow:	movlw	(1<<FRONT|1<<REAR)
	movwf	ibar
	call	makeneutral
	; then do the sync part. also send the pulses a bit faster than normal
	movlw	.18
	call	delayXms
	decfsz	foo,F
	goto	doshow

	; give control back to the user and start with the default setup
	movlw	1
	movwf	opmode
	clrf	mstate		; forget old mswitch states
	clrf	prevmstate	; and the possibly dud 
	movfw	GPIO		; read latch
	bcf	INTCON,RAIF	; clear pinchange
	bsf	INTCON,RAIE	; re-enable pinchange int

	btfsc	GPIO,LSWITCH	; is the lockswitch jumpered?
	goto	preploop	; no, continue.

	movlw	2		; fixed fws (2ws-front)
	movwf	opmode
	goto	$		; and forget about switching modes.

preploop:
 	; we ran without interrupts and messed things up, so wait a little to get correct mswitch readings
	movlw	.50
	call	delayXms

	; modebeep if cyclic, long beep for direct
	btfss	config,CYCBIT
	call	bleep		; !cyc=direct mode
	call	modebeep	; uses bar, doesn't beep in direct mode
modeloop:
	movlw	.50
	call	delayXms	; once every .05s should be pretty much fast enough
	call	waitmode	; also wait until the mode timer is not running, max extra delay 2ms

	call	decodestate	; don't do anything if there is no valid signal
	movwf	mstate
	btfss	mstate,VALBIT	; mswitch dud? do nothing.
	goto	notrans

	; direct mode or cyclic?
	btfsc	config,CYCBIT
	goto	check_cyclic

	; we are in direct mode.
	; mswitch levels are below neutral-deadband/above neutral+deadband and the rest, ie. neutral
	; xlat and d2mtable wants 0 high, 1 neut, 2 low
	movlw	1		; for neutral
	btfsc	mstate,DBBIT	; in neutral range?
	goto	xlat
	movlw	2		; that's low
	btfss	mstate,HIBIT	; skip if high
	goto	xlat
	movlw	0

	; determine what that mswitch position modulo DM0BIT/DM1BIT settings means mode-wise
xlat:	btfsc	config,DM1BIT
	addlw	3
	call	d2mtable
	movwf	foo
	btfsc	config,DM0BIT
	swapf	foo,F		; high nibble counts if dm0bit is set
	movlw	0x0f
	andwf	foo,W		; and there's our mode.
	
	; desired opmode same as before? leave it where it is.
	movwf	foo
	xorwf	opmode,W	; same?
	skpnz
	goto	notrans
	movfw	foo		; new opmode is different
	goto	setmode

	; we are in cyclic mode, change on some type of transition
check_cyclic:
	btfsc	config,TRNBIT	; want double transitions?
	goto	dbltrans
	
	btfss	mstate,HIBIT	; skip if high now
	btfss	prevmstate,HIBIT	; skip if prev was high
	goto	notrans
	goto	incmod

dbltrans:
	movfw	transcount	; read the transition counter
	skpz			; and decrement it only if it's still fine
	decf	transcount,F	; one less attempt to find a suitable transition
	movwf	foo		; temp storage

	bcf	prevmstate,DBBIT ; we don't care about the deadband 
	movfw	mstate
	andlw	~(1<<DBBIT)	; get rid of the deadband bit for the comparison
	xorwf	prevmstate,W	; nonzero: a transition, either way.
	skpnz
	goto	notrans
	; ok, a transition. let's reset the counter.
	movlw	MAXDELTA
	movwf	transcount
	; and was the delay between transitions short enough?
	movfw	foo
	skpnz			; zero:	nope, count ran out - start from scratch
	goto	notrans
	clrf	transcount	; but a mode-switching transition doesn't start a cycle
	
incmod:	; let's switch modes
	incf	opmode,W
	movwf	foo		; no messing with opmode until we're finished.
	; skip any disabled modes
	addlw	-3		; mode 3, rws
	skpz
	goto	checkm4
	btfss	config,RSCBIT	; mode 3 included or not?
	incf	foo,F		; not.
checkm4:
	movlw	-4		; mode 4, crab
	addwf	foo,W
	skpz
	goto	checkend
	btfss	config,CRCBIT	; mode 4 included or not?
	incf	foo,F		
checkend:		
	movlw	5		
	subwf	foo,W
	movfw	foo		; doesn't alter carry
	skpnc			; skip if subtraction gave ~carry=borrow, ie. opmode-5 is negative
	movlw	1		; skipped if <=4

setmode:
	; wait until the steering timer is not running
	; (so that we don't mess a current output operation)
	btfsc  	T1CON,TMR1ON	
	goto	$-1
	movwf	opmode
	call	modebeep	; uses bar
notrans:	
	movfw	mstate		; remember, this is way later and mswitch might have changed inbetween
	movwf	prevmstate	; so we save the value we worked on
	goto	modeloop

; delay until mode is valid		
waitmode: 	; wait until the mode timer is not running
	BANK1	
	btfsc  	OPTION_REG,T0CS	
	goto	$-1
	BANK0
	return

; N short beeps (as per opmode), but no beeping in direct mode
; trashes W, uses bar
modebeep:
	btfss	config,CYCBIT
	return
	movfw	opmode
	movwf	bar
onemore:
	call	blip
	call	delay100ms
	decfsz	bar,F
	goto	onemore
	return

; w holds bit to sound. blip for OFF, blip-blip for ON
; destroys w, uses bar
soundstate:
	clrf	bar
	incf	bar,F		; one blip for sure
	addlw	0		; test for zero
	skpz
	incf	bar,F		; two if on
	goto	onemore

; makes one 50ms blip, no trailing pause
; destroys w
blip:	bsf	GPIO,BEEP
	movlw	.50
	call	delayXms
	bcf	GPIO,BEEP
	return

; makes one 256ms long beep, no trailing pause
; destroys w
bleep:	bsf	GPIO,BEEP
	clrw
	call	delayXms
	bcf	GPIO,BEEP
	return

	
; nonrecoverable fault:	beeper on 4s, then sleep
fault:	bcf	INTCON,GIE	; let's not get interrupted :-)
	bsf	GPIO,BEEP
	movlw	.16		; 16x256ms = 4096ms
	movwf	foo
more:	clrw
	call	delayXms	; 256ms
	decfsz	foo,F
	goto	more
	bcf	GPIO,BEEP
	sleep

; two main scenarios: port int-on-change for steering or modeswitch,
; or tmr1 expiration when making output pulses
; runs in bank0
inthandler:
	; is it a tmr1 overflow event? check only if the int is enabled	
	BANK1
	btfss	PIE1,TMR1IE	; check if timer int is enabled
	goto	pin_test	; skipped if set, timer-related
	BANK0
	btfss	PIR1,TMR1IF	; timer overflow?
	goto	pin_test	; skipped if set

	; do the timer overflow handling then: making a pulse has ended
	bcf	T1CON,TMR1ON	; stop timer1
	bcf	PIR1,TMR1IF	; ack the interrupt
	BANK1
	bcf	PIE1,TMR1IE	; disable the tmr1 interrupt
	BANK0

	; lazy:	we switch off both outputs regardless of mode
	; works because the other is definitely off anyway 
	; as we're in the sync delay between pulses
	movlw	~(1<<FRONT|1<<REAR)
	andwf	GPIO,F
	
	; in case of mode 3 + reversed rear we need to make two
	; computed pulses, first the reversed signal on rear,
	; then the neutral on front
	btfss	lateneutral,0
	goto	pin_test
	; set this late neutral up, once (steering signal trigger will set flag again)
	clrf	lateneutral
	call	neutmodes
	movwf	ibar
	call	makeneutral
	
	; or an int-on-change?
pin_test:
	BANK0
	btfss	INTCON,RAIF	; pin-change interrupt ? (skip if set)
	retlw	0		; not a pin interrupt, we ignore it

	; determine whether steering or modeswitch pins changed
	; start appropriate action.
	movfw	GPIO		; save and (clear port mismatch state)
	movwf	ifoo
	bcf	INTCON,RAIF	; clear int condition

	movfw	ifoo	
	xorwf	lastpins,F	; which pins have changed state?
	btfss	lastpins,STEER
	goto	check_mswitch	; skipped if steering

	; a steering pulse, up or down?
	btfss	ifoo,STEER
	goto	steer_end ; skipped if up

	; up: relevant outputs on, start timer countdown
	movlw	~(1<<FRONT|1<<REAR)
	andwf	GPIO,F		; outputs off first, to be sure

	clrf	lateneutral	; special-case handling for mode3 + reversed rear channel
	tstf	opmode		; opmode off? no pulses then, ever.
	skpnz
	goto	nopulse
	call	directmodes
	movwf	ibar		; save
	tstf	ibar
	skpnz			; all normal modes have *some* direct output, sole non-direct mode is 3+rev
	bsf	lateneutral,0	; record that a late neutral is to follow

	movfw	ibar		; and restore
	iorwf	GPIO,F		; relevant outputs on again
nopulse:	
	clrf	TMR1L
	clrf	TMR1H
	bsf	T1CON,TMR1ON	; timer on!	
	goto	check_mswitch
steer_end:
	; down:	 stop output front/rear, stop and save timer
	movlw	~(1<<FRONT|1<<REAR)
	andwf	GPIO,F		; outputs off first, to be sure
	bcf	T1CON,TMR1ON	; timer off
	movfw	TMR1L
	movwf	steerlo
	movfw	TMR1H
	movwf	steerhi

	; in opmode 0, programming? no pulses
	tstf	opmode
	skpnz
	goto	check_mswitch
	
	; now check if any inverted pulses need to be output
	call	revmodes
	movwf	ibar		; remember which channel to reverse on, signal inversion needs W
	addlw	0		; to check for zero
	skpnz			; no inverted signal wanted? then look if neutrals are to be done
	goto	checkneutral

	; make inverted pulse
	; delta is neutral-current, want neutral+delta,
	; or more precisely tmr1 preload value 0xffff-(neutral+delta) 
	; =0xffff-(2*neutral-current)
	bcf	STATUS,C	; rotate through carry
	rlf	neutlo,W	; neut*2, into tmr1
	movwf	TMR1L
	rlf	neuthi,W	; carry is carried over
	movwf	TMR1H

	movfw	steerlo
	subwf	TMR1L,F
	skpc			
	decf	TMR1H,F		; skipped if carry = ~borrow
	movfw	steerhi
	subwf	TMR1H,F
	comf	TMR1H,F		; unclean but sufficiently precise
	comf	TMR1L,F
	call	dopulse		; uses ibar
	goto	check_mswitch

checkneutral:
	call	neutmodes	; any neutral signals required?
	movwf	ibar
	addlw	0		; check for zero
	skpz
	call	makeneutral
	
check_mswitch:
	btfss	lastpins,MSWITCH
	goto	fin
	; a mswitch pulse, up or down?
	btfss	ifoo,MSWITCH
	goto	mswitch_end ; skipped if up
	; up: clear and start tmr0
	clrf	TMR0
	BANK1
	bcf	OPTION_REG,T0CS
	BANK0
	goto	fin
mswitch_end:
	; down:	stop and save tmr0
	BANK1
	bsf	OPTION_REG,T0CS
	BANK0
	movfw	TMR0
	movwf	mswitch
	
fin:	movfw	ifoo
	movwf	lastpins	; save new pin state
	retlw	0

; makes neutral pulse with gpio latch stored in ibar
; note:	second entry point dopulse
makeneutral:
	comf	neuthi,W	; unclean but sufficiently precise
	movwf	TMR1H
	comf	neutlo,W
	movwf	TMR1L
dopulse:
	movfw	ibar		; the channels we want the tmr1-defined pulse output
	bcf	PIR1,TMR1IF	; clear int state for safety
	BANK1
	bsf	PIE1,TMR1IE	; enable tmr1 int
	BANK0
	iorwf	GPIO,F		; pulse!
	bsf	T1CON,TMR1ON	; and timer on
	return

; converts current switch channel pulse length to trinary switch state
; input in mswitch:  pulse length
; returns in W:	b0: 1 valid, b1: 1 above center, b2: 1 in deadband
; uses tempvars baz 
decodestate:
	movfw	mswitch		; mswitch is interrupt-controlled
	movwf	baz		; temp save current value
	; first a sanity check: above min pulse length, below max pulse length
	movlw	(MAXPULSE/.16)
	subwf	baz,W		; should to be negative, ie borrow happened
	skpnc			; ~carry=borrow, skips if neg
	retlw	0		; above max
	movlw	(MINPULSE/.16)
	subwf	baz,W		; should be positive, ie no borrow
	skpc			; carry=~borrow, skips if pos
	retlw	0
	; below neutral-deadband or above neutral+deadband?
	movlw	(CTRPULSE+DEADBAND)/.16 ; above neutral+dead?
	subwf	baz,W
	skpnc			; borrow = ~carry = negative, skipped if neg
	retlw	(1<<HIBIT|1<<VALBIT) ; 3=high and outside deadband
	movlw	(CTRPULSE-DEADBAND)/.16 ; below neutral-dead?
	subwf	baz,W
	skpc			; ~borrow = carry, ie. positive, skips if pos
	retlw	(1<<VALBIT)	; 1=low and ouside deadband
	; neutral it is, but below or above center?
	movlw	(CTRPULSE/.16)
	subwf	baz,W
	skpnc			; skips if negative
	retlw	(1<<VALBIT|1<<HIBIT|1<<DBBIT) 	; 7=high but inside deadband
	retlw	(1<<VALBIT|1<<DBBIT) 	; 5=low but inside deadband 

#include "delay.inc"

	org	0x2000		; id words
	org	0x2100		; eeprom
	de	0		; permanent storage for the configword 
	de	"(c) 2008-2010 az fourws $Revision: 1.11 $"
end